/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include "chainmanager.hpp"
#include <memory>
#include <sys/select.h>
#include <sched.h>
#include <algorithm>
#include <sys/time.h>

#if ENABLE_DEBUG
#include <iostream>
#endif

namespace flux
{
    ChainManager::ChainManager()
    {
    }
    ChainManager::~ChainManager()
    {
        for (ChainsMap::iterator chain_i = chains_.begin(); chain_i != chains_.end(); ++chain_i)
            delete chain_i->second;
    }

    bool ChainManager::registerInput(io::IOChecker checker, RegInfo<io::Input> const &info)
    {
        if (input_types_.find(checker) != input_types_.end())
            return false;

        input_types_[checker] = info;
        return true;
    }

    bool ChainManager::registerOutput(io::IOChecker checker, RegInfo<io::Output> const &info)
    {
        if (output_types_.find(checker) != output_types_.end())
            return false;

        output_types_[checker] = info;
        return true;
    }

    bool ChainManager::registerParser(std::string const &name, RegInfo<parse::Parser> const &info)
    {
        if (parser_names_.find(name) != parser_names_.end())
            return false;

        parser_names_[name] = info;
        return true;
    }

    std::string ChainManager::createChain(std::list<std::string> const &input,
                                          std::string const &parser,
                                          std::list<std::string> const &output,
                                          std::string const &suggested_id)
    {
        RefCounted<io::Input> in;
        RefCounted<io::Output> out;
        std::auto_ptr<parse::Parser> p;
        std::auto_ptr<Chain> c;

        std::string id;

        //Parse parser name and options
        std::string::size_type parser_opts_pos = parser.find_first_of(":");
        std::string parser_name = parser.substr(0, parser_opts_pos);
        std::string parser_opts;
        if (parser_opts_pos != std::string::npos)
            parser_opts = parser.substr(parser_opts_pos + 1);

        //Lookup if we know the requested parser
        ParserNamesMap::iterator parser_i = parser_names_.find(parser_name);
        if (parser_i == parser_names_.end())
            return "<Invalid Parser>";
        else
            p.reset(parser_i->second.allocator(parser_opts));

        id += parser_name;
        c.reset(new Chain(p.release()));

        for (std::list<std::string>::const_iterator input_i = input.begin(); input_i != input.end(); ++input_i) {
            std::string const &cur_input = *input_i;
#if ENABLE_DEBUG
            std::cerr << "Chain IN: " << cur_input << std::endl;
#endif

            //Lookup if we already created an input for the specified input
            AllocatedInputMap::iterator inputs_end = inputs_.end();
            for (AllocatedInputMap::iterator i = inputs_.begin(); i != inputs_end; ++i) {
                if (i->first == cur_input) {
                    in = i->second;
                    break;
                }
            }

            //If we didn't find a preallocated input and we have an Input allocator create it
            if (!in) {
                InputTypesMap::iterator input_types_end = input_types_.end();
                for (InputTypesMap::iterator i = input_types_.begin(); i != input_types_end; ++i) {
                    if (i->first(cur_input)) {
                        in = i->second.allocator(cur_input);
                        inputs_[cur_input] = in;
                        break;
                    }
                }
            }

            if (!in)
                return "<Invalid Input>";

            c->inputs.push_back(in);
            id += cur_input;
        }

        for (std::list<std::string>::const_iterator output_i = output.begin(); output_i
            != output.end(); ++output_i) {
            std::string const &cur_output = *output_i;
#if ENABLE_DEBUG
            std::cerr << "Chain OUT: " << cur_output << std::endl;
#endif

            //Look if we have an allocator for the requested output
            OutputTypesMap::iterator output_types_end = output_types_.end();
            for (OutputTypesMap::iterator i = output_types_.begin(); i != output_types_end; ++i) {
                if (i->first(cur_output)) {
                    out = i->second.allocator(cur_output);
                    break;
                }
            }

            if (!out)
                return "<Invalid Output>";

            c->outputs.push_back(out);
            id += cur_output;
        }

        if (!suggested_id.empty())
            id = suggested_id;

        chains_[id] = c.release();
        return id;
    }

    void ChainManager::cycle()
    {
        ChainsMap::iterator chains_end = chains_.end();
        AllocatedInputMap::iterator inputs_end = inputs_.end();

        fd_set readset;
        int max_fd = -1;
        struct timeval timeout = { 1, 0 };

        //Block until some data is available
        FD_ZERO(&readset);
        for (AllocatedInputMap::iterator input_i = inputs_.begin(); input_i != inputs_end; ++input_i) {
            io::Input *in = input_i->second.ptr();
            if (in->isAvailable()) {
                max_fd = std::max(in->getDescriptor(), max_fd);
                FD_SET(in->getDescriptor(), &readset);
            }
            else
                in->recover();
        }

        if (max_fd == -1) {
            sched_yield();
            return;
        }

        select(max_fd + 1, &readset, NULL, NULL, &timeout);

        //Read data from inputs
        for (AllocatedInputMap::iterator input_i = inputs_.begin(); input_i != inputs_end; ++input_i) {
            io::Input *in = input_i->second.ptr();
            if (FD_ISSET(in->getDescriptor(), &readset))
                in->read();
            else
                in->feed();
        }

        for (ChainsMap::iterator chain_i = chains_.begin(); chain_i != chains_end; ++chain_i) {
            Chain *chain = chain_i->second;

            for (Chain::InputList::iterator input = chain->inputs.begin();
                 input != chain->inputs.end(); ++input) {
                io::Input *in = input->ptr();
                chain->parser->parse(in->getPacket());
            }

            for (Chain::OutputList::iterator output = chain->outputs.begin();
                 output != chain->outputs.end(); ++output) {
                 io::Output *out = output->ptr();
                do {
                    if (out->isAvailable())
                        out->write(chain->parser->getResult(), chain->parser->getResultSize());
                    else
                        out->recover();
                }
                while (chain->parser->hasMore());
            }
        }
    }

}

#ifdef TEST
#include "../io/io_udp.hpp"
#include <iostream>

using namespace flux;

int main()
{
    flux::ChainManager cm;
    std::list<std::string> in;
    std::list<std::string> out;

    in.push_back("udp://1900");
    in.push_back("udp://1902");
    out.push_back("udp://localhost:9000");
    out.push_back("udp://localhost:9002");

    if (!cm.createChain(in, "copy", out).empty()) {
        std::cerr << "Have been able to create a chain without any registered Input" << std::endl;
        return 5;
    }

    cm.registerInput(flux::io::UDPInput::Check, ChainManager::RegInfo<io::Input>(io::UDPInput::Alloc, "desc"));
    if (cm.registerInput(flux::io::UDPInput::Check, ChainManager::RegInfo<io::Input>(io::UDPInput::Alloc, "desc"))) {
        std::cerr << "Have been able to reassign a slot to an already registered Input checker" << std::endl;
        return 1;
    }

    if (!cm.createChain(in, "copy", out).empty()) {
        std::cerr << "Have been able to create a chain without any registered Output" << std::endl;
        return 6;
    }

    cm.registerOutput(flux::io::UDPOutput::Check, ChainManager::RegInfo<io::UDPOutput>(io::UDPOutput::Alloc, "desc"));
    if (cm.registerOutput(flux::io::UDPOutput::Check, ChainManager::RegInfo<io::UDPOutput>(io::UDPOutput::Alloc, "desc")) {
            std::cerr << "Have been able to reassign a slot to an already registered Output checker" << std::endl;
            return 2;
        }

        if (!cm.createChain(in, "copy", out).empty()) {
            std::cerr << "Have been able to create a chain without any registered Parser" << std::endl;
            return 7;
        }

        cm.registerParser("copy", ChainManager::RegInfo<parse::Parser>(parse::CopyParser::Alloc, "desc"));
        if (cm.registerParser("copy", ChainManager::RegInfo<parse::Parser>(parse::CopyParser::Alloc, "desc"))) {
            std::cerr << "Have been able to reassign a slot to an already registered parser" << std::endl;
            return 3;
        }

        if (cm.createChain(in, "copy", out).empty()) {
            std::cerr << "Unable to create a chain" << std::endl;
            return 4;
        }
        std::cout << "+ CHAINMANAGER TEST SUITE PASSED" << std::endl;
        return 0;
    }

#endif

