/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

static char const * const Version = "1.1";

#include <string>
#include <list>
#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include "signal.h"
#include <cstring>

#include "chain/chainmanager.hpp"

#include "io/io_udp.hpp"
#include "io/io_tcp.hpp"
#include "io/io_mq.hpp"
#include "io/io_file.hpp"

#ifdef BUFFERPOOL
# include "io/io_bufferpool.hpp"
#endif

#include "parser/parser.hpp"
#include "parser/live.hpp"

#ifdef RTP
#  include "parser/unpack.hpp"
#endif

#ifdef TRANSCODE
#  include "parser/transcode.hpp"
#endif

#ifdef XMLRPC
#       include "utils/rpc.hpp"
RpcServer rpcserver;
Sock *rpcsock;
#endif

int stop = 0;
flux::ChainManager *cm = 0;

static void terminator_function(int num)
{
  std::cout << "Exiting..." << std::endl;

#ifdef XMLRPC
  rpc_close(&rpcserver, rpcsock);
#endif

  if (cm)
    delete cm;

  exit(0);
}

static void handle_signals()
{
    struct sigaction term_action;
    sigset_t block_set;

    /* catch TERM, INT & BUS signals */
    memset(&term_action, 0, sizeof(term_action));
    term_action.sa_handler = terminator_function;
    sigaction(SIGINT, &term_action, NULL);
    sigaction(SIGTERM, &term_action, NULL);
    sigaction(SIGBUS, &term_action, NULL);

    /* block PIPE signal */
    sigemptyset(&block_set);
    sigaddset(&block_set, SIGPIPE);
    sigprocmask(SIG_BLOCK, &block_set, NULL);
}

void show_help(flux::ChainManager *manager)
{
    flux::ChainManager::InputTypesMap const &inputs =  manager->getInputs();
    flux::ChainManager::OutputTypesMap const &outputs = manager->getOutputs();
    flux::ChainManager::ParserNamesMap const &parsers = manager->getParsers();

    std::cout << "Flux (LSCube Streaming Suite) " << Version << std::endl;
    std::cout << "Copyright (C) 2008-2009 LScube team <team@lscube.org>" << std::endl;
    std::cout << "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>" << std::endl;
    std::cout << std::endl;
    std::cout << "Usage: flux [OPTIONS]" << std::endl;
    std::cout << "\t -h            Display this text" << std::endl;
    std::cout << "\t -c <chain>    Specify a work chain" << std::endl;
    std::cout << std::endl;
    std::cout << "Chain Description:" << std::endl;
    std::cout << "\t A Chain is a group of inputs and outputs joined by a parser." << std::endl;
    std::cout << "\t in the form: INPUT[,INPUT2[...]]*PARSER:PARSER_OPTIONS*OUTPUT[,OUTPUT2[...]]" << std::endl;
    std::cout << "\t The three entities are separated by an asterisk (*) and each" << std::endl;
    std::cout << "\t input or ouput is separated by a comma (,) from its siblings" << std::endl;
    std::cout << "\t some parsers might also take options, refer to their documentation" << std::endl;
    std::cout << "\n\t EXAMPLE: -c udp://1900*copy*udp://server1.org:1900,udp://server2.org:1900" << std::endl;
    std::cout << std::endl;

    std::cout << "Available Inputs:" << std::endl;
    for (flux::ChainManager::InputTypesMap::const_iterator it = inputs.begin(); it != inputs.end(); ++it)
        std::cout << "\t" << it->second.description << std::endl;
    std::cout << std::endl;

    std::cout << "Available Parsers:" << std::endl;
    for (flux::ChainManager::ParserNamesMap::const_iterator it = parsers.begin(); it != parsers.end(); ++it)
        std::cout << "\t" << it->second.description << std::endl;
    std::cout << std::endl;

    std::cout << "Available Outputs:" << std::endl;
    for (flux::ChainManager::OutputTypesMap::const_iterator it = outputs.begin(); it != outputs.end(); ++it)
        std::cout << "\t" << it->second.description << std::endl;
    std::cout << std::endl;
}

//Implemented in chain/chain_parse.rl
bool chain_parse(std::string const &text, std::list<std::string> &inputs, std::string &parser, std::list<std::string> &outputs);

using namespace flux;

int main(int argc, char *argv[])
{
  cm = new flux::ChainManager();
  handle_signals();

  cm->registerInput(io::UDPInput::Check, ChainManager::RegInfo<io::Input>(io::UDPInput::Alloc, "udp://port - UDP Input"));
  cm->registerInput(io::TCPInput::Check, ChainManager::RegInfo<io::Input>(io::TCPInput::Alloc, "tcp://port - TCP Input"));
  cm->registerInput(io::FileInput::Check, ChainManager::RegInfo<io::Input>(io::FileInput::Alloc, "file://path - Read from a file"));

  cm->registerParser("copy", ChainManager::RegInfo<parse::Parser>(parse::CopyParser::Alloc, "copy - Plain Copy of packets"));
  cm->registerParser("live", ChainManager::RegInfo<parse::Parser>(parse::LiveParser::Alloc, "live - Copy and sync timestamps and SSRC"));

#ifdef RTP
  cm->registerParser("unpack", ChainManager::RegInfo<parse::Parser>(parse::UnpackParser::Alloc, "unpack - Demux from RTP the stream"));
#endif

#ifdef TRANSCODE
  cm->registerParser("transcode", ChainManager::RegInfo<parse::Parser>(parse::TranscodeParser::Alloc, "transcode - Demux from RTP and transcode to mpeg2video (might have :WIDTHxHEIGHT option to scale the video)"));
  cm->registerParser("transcodepack", ChainManager::RegInfo<parse::Parser>(parse::TranscodeParser::PackAlloc, "transcodepack - Demux from RTP and transcode to mpeg2video and remux to RTP (might have :WIDTHxHEIGHT option to scale the video)"));
  cm->registerParser("elive", ChainManager::RegInfo<parse::Parser>(parse::EmptyLiveParser::Alloc, "elive - Like live parser, but will place black frames after a timeout"));
#endif

  cm->registerOutput(flux::io::UDPOutput::Check, ChainManager::RegInfo<io::Output>(io::UDPOutput::Alloc, "udp://host:port - UDP Output"));
  cm->registerOutput(flux::io::TCPOutput::Check, ChainManager::RegInfo<io::Output>(io::TCPOutput::Alloc, "tcp://host:port - TCP Output"));
  cm->registerOutput(flux::io::FileOutput::Check, ChainManager::RegInfo<io::Output>(io::FileOutput::Alloc, "file://path - File Output"));
 cm->registerOutput(flux::io::MQOutput::Check, ChainManager::RegInfo<io::Output>(io::MQOutput::Alloc, "mq://path - MQ Output"));
#ifdef BUFFERPOOL
  cm->registerOutput(flux::io::BufferpoolOutput::Check, ChainManager::RegInfo<io::Output>(io::BufferpoolOutput::Alloc, "bufferpool://shm.rtp - Feng Shared Memory output"));
#endif

  if (argc <= 1) {
    show_help(cm);
    return 0;
  }

  for (int i=1; i<argc; ++i) {
    if ( (argv[i][0] == '-') && (argv[i][1] != '\0') ) {
      switch(argv[i][1]) {
        case 'c':
            if ( ((i+1) < argc) ) {
              std::string chain_def = argv[++i];
              std::list<std::string> ins;
              std::list<std::string> outs;
              std::string parser;

              std::cout << "Requested Chain: " << chain_def << std::endl;
              if (chain_parse(chain_def, ins, parser, outs)) {
                try {
                  std::cout << "Created Chain: " << cm->createChain(ins, parser, outs) << std::endl;
                }
                catch (std::exception &e) {
                  std::cerr << "Unable to create chain: " << e.what() << std::endl;
                }
              }
              else
                std::cerr << "Unable to recognize Chain" << std::endl;
            }
          break;
        case 'h':
          show_help(cm);
          return 0;
        default:
          std::cerr << "Unrecognized option -" << argv[i][1] << std::endl;
          break;
      }
    }
  }

#ifdef XMLRPC
  rpc_init(cm, &rpcserver, &rpcsock);
#endif

  while(!stop) {
#ifdef XMLRPC
    if (!stop)
      rpc_handle_req(&rpcserver, rpcsock);
#endif
    if (!stop)
      cm->cycle();
  }

  std::cout << "Cleaning Up" << std::endl;
  terminator_function(0);

  return 0;
}
