/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include "io_file.hpp"

#include <unistd.h>
#include <iostream>
#include <fstream>
#include <stdexcept>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace flux
{
    namespace io
    {
        static int OpenFileForInput(char const *uri)
        {
            if ((uri[0] == '-') && (uri[1] == '\0')) {
                return 0;
            }
            else {
                int fd = open(uri, O_RDONLY);
                if (fd == -1)
                    throw std::runtime_error("Unable to open FileInput");
                return fd;
            }
        }

        FileInput::FileInput(char const *uri) :
            Input(uri)
        {
            recover();
        }

        FileInput::~FileInput()
        {
            if (isAvailable())
                close(fd_);
        }

        bool FileInput::isAvailable()
        {
            return fd_ != -1;
        }

        bool FileInput::recover()
        {
            fd_ = OpenFileForInput(uri_.c_str());
            return isAvailable();
        }

        void FileInput::feed()
        {
            received_packet_.parseBuffer(0);
        }

        unsigned long FileInput::read()
        {
            long long size = 0;

            size = ::read(fd_, received_packet_.getBuffer(), 10);
            if (size <= 0)
                size = 0;

            received_packet_.parseBuffer(size); //parse it just to save size

            return size;
        }

        bool FileInput::Check(std::string const &uri)
        {
            return (uri.find("file://") == 0);
        }

        Input *FileInput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("file://") + 7);
            return new FileInput(new_uri.c_str());
        }

        rtp::Packet &FileInput::getPacket()
        {
            return received_packet_;
        }

        static std::ofstream *OpenFileForOutput(char const *uri)
        {
            std::ofstream *f = new std::ofstream(uri, std::ios::out | std::ios::binary);
            if (!f->is_open()) {
                std::cerr << "Unable to open output file " << uri << std::endl;
                delete f;
                return 0;
            }
            else
                return f;
        }

        FileOutput::FileOutput(char const *uri) :
            Output(uri), output_(OpenFileForOutput(uri))
        {
        }

        FileOutput::~FileOutput()
        {
            if (output_)
                output_->close();

            delete output_;
        }

        bool FileOutput::isAvailable()
        {
            return ((output_ != 0) && (output_->is_open()) && (output_->good()));
        }

        bool FileOutput::recover()
        {
            delete output_;
            output_ = OpenFileForOutput(uri_.c_str());
            return isAvailable();
        }

        unsigned long FileOutput::write(unsigned char *buffer, unsigned long buffer_size)
        {
#if ENABLE_DEBUG
            std::cerr << "Dumping " << buffer_size << " bytes" << std::endl;
#endif

            if (!buffer_size)
                return 0;

            output_->write(reinterpret_cast<const char*> (buffer), buffer_size);
            output_->flush();

            return buffer_size;
        }

        bool FileOutput::Check(std::string const &uri)
        {
            return (uri.find("file://") == 0);
        }

        Output *FileOutput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("file://") + 7);
            return new FileOutput(new_uri.c_str());
        }
    }
}

#ifdef TEST
#include <iostream>

int main()
{
    flux::io::FileOutput out("test.out");

    if (!out.isAvailable()) {
        std::cerr << "FileOutput not opened correctly" << std::endl;
        return 1;
    }

    //Test packet transmission
    flux::rtp::Packet p;
    p.setVersion(2);
    p.setSSRC(12345);
    p.regenBuffer();

    out.write(p.getBuffer(), p.getBufferSize());

    if (!out.isAvailable()) {
        std::cerr << "FileOutput not available after write" << std::endl;
        return 2;
    }

    std::cout << "+ FILE I/O TEST SUITE PASSED" << std::endl;
    return 0;
}
#endif

