/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include "io_mq.hpp"

#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>

#define SEQ_TRESHOLD 5

#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <cstring>

#include "../rtp/packet.hpp"

namespace flux
{
  namespace io
  {
    static mqd_t OpenForOutput(char const *uri)
    {
        mq_attr attr;
        attr.mq_maxmsg = MQOutput::MaxQueueMessages;
        attr.mq_msgsize = MQOutput::QueueMessageSize;
        attr.mq_flags = 0;
        mqd_t mpd = mq_open(uri, O_CREAT|O_RDWR, S_IRWXU, &attr);

        if (mpd < 0)
            std::cerr << "MQOutput could not create queue '" << uri << "' ("  << strerror(errno) << ")" << std::endl;
        return mpd;
    }

    MQOutput::MQOutput(char const *uri)
    : Output(uri), mqd_(OpenForOutput(uri)), last_seq_(-1), seq_stray_(0)
    {
    }

    MQOutput::~MQOutput()
    {
        if (mqd_>=0)
            mq_close(mqd_);

        mq_unlink(uri_.c_str());
    }

    bool MQOutput::isAvailable()
    {
        return (mqd_ >= 0);
    }

    bool MQOutput::recover()
    {
        if (isAvailable())
            mq_close(mqd_);

        mqd_ = OpenForOutput(uri_.c_str());

        return isAvailable();
    }

    static inline bool seq_restarted(unsigned last_seq_, unsigned cur_seq_, unsigned seq_stray_)
    {
        if (seq_stray_ != 0)
            return false;

        if ( (cur_seq_ == 0) && (last_seq_ == 65535) )
            return true;

        return false;
    }

    unsigned long MQOutput::write(unsigned char *buffer, unsigned long buffer_size)
    {
        if (!buffer_size)
            return 0;

        mq_attr attr;
        mq_getattr(mqd_, &attr);
        if (attr.mq_curmsgs >= MaxQueueMessages) {
            std::cerr << "[MQ:" << uri_ << "] full! trashing..." << std::endl;
            return 0;
        }

        rtp::Packet p;
        memcpy(p.getBuffer(), buffer, buffer_size);
        p.parseBuffer(buffer_size);

        if (last_seq_ < 0) {
            std::cerr << "[MQ:" << uri_ << "] Rebased!" << std::endl;

            last_seq_ = p.getSeq();
            prev_ts_ = p.getTimestamp();
            delivery_ts_ = 0;
        }
        else if ((abs(p.getSeq() - last_seq_) > SEQ_TRESHOLD) && !seq_restarted(last_seq_, p.getSeq(), seq_stray_)) {

            std::cerr << "[MQ:" << uri_ << "] " <<
                         "Packet with previous/far sequence number " <<
                         abs(p.getSeq() - last_seq_) << " (" <<
                         p.getSeq() << " - " << last_seq_ << ")" << std::endl;

            ++seq_stray_;
            if (seq_stray_ >= SEQ_TRESHOLD) {
                MQMetaData::Rebase(this);
                last_seq_ = -1;
                seq_stray_ = 0;
            }

            return 0;
        }

        last_seq_ = p.getSeq();
        delivery_ts_ += (p.getTimestamp() - prev_ts_);
        prev_ts_ = p.getTimestamp();
        meta_.rebuild(this, delivery_ts_);

        char outbuffer[QueueMessageSize];
        memcpy(outbuffer, meta_.getData(), meta_.getSize());
        memcpy(outbuffer+meta_.getSize(), (char*)p.getBuffer(), p.getBufferSize());

        if (mq_send(mqd_, outbuffer, p.getBufferSize()+meta_.getSize(), 0) <0)
            std::cerr << "[" << uri_ << "] error in MQ writing" << std::endl;

        return buffer_size;
    }


    bool MQOutput::Check(std::string const &uri)
    {
      return (uri.find("mq://") == 0);
    }

    Output *MQOutput::Alloc(std::string const &uri)
    {
      std::string new_uri = uri.substr(uri.find("mq://")+5);
      return new MQOutput(new_uri.c_str());
    }
  }
}
