/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include <stdexcept>
#include <iostream>
#include <cstring>
#include <time.h>

extern "C"
{
#include <netembryo/wsocket.h>
}

#include "io_tcp.hpp"

const unsigned FLUX_TCP_TIMEOUT = 5;

static void set_socket_timeout(Sock * s, int type)
{
    struct timeval timeout = { FLUX_TCP_TIMEOUT, 0 };
    if (setsockopt(Sock_fd(s), SOL_SOCKET, type, (void *) &timeout, sizeof(timeout)) == -1)
        std::cerr << "Unable to set socket timeout" << std::endl;
}

namespace flux
{
    namespace io
    {
        TCPInput::TCPInput(char const *uri) :
            Input(uri), listen_socket_(0), transport_(0), ok_(false), last_pkt_time_(0)
        {
            if (!is_a_port(uri_))
                throw std::invalid_argument(
                                            "TCPInput requires a valid even port number as parameter");

            listen_socket_ = Sock_bind(NULL, uri_.c_str(), NULL, TCP, 0);
            Sock_listen(static_cast<Sock*> (listen_socket_), 1);
            recover();
        }

        TCPInput::~TCPInput()
        {
            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));

            if (listen_socket_)
                Sock_close(static_cast<Sock*> (listen_socket_));
        }

        bool TCPInput::isAvailable()
        {
            return ok_;
        }

        bool TCPInput::recover()
        {
            ok_ = false;
            fd_ = -1;

            if (transport_) {
                Sock_close(static_cast<Sock*> (transport_));
                transport_ = 0;
            }

            if (!listen_socket_)
                throw std::runtime_error("TCPInput listen socket is invalid");

            transport_ = Sock_accept(static_cast<Sock*> (listen_socket_), NULL);
            if (!transport_)
                throw std::runtime_error("Unable to recover TCPInput " + uri_);
            set_socket_timeout(static_cast<Sock*> (transport_), SO_RCVTIMEO);

            fd_ = Sock_fd(static_cast<Sock*>(transport_));
            last_pkt_time_ = time(0);
            ok_ = true;
            return ok_;
        }

        void TCPInput::feed()
        {
            received_packet_.parseBuffer(0);
        }

        unsigned long TCPInput::read()
        {
            uint32_t tcp_size = 0;
            long size;

            received_packet_.parseBuffer(0);
            if (!ok_)
                return 0;

            if ((time(0) - last_pkt_time_) >= FLUX_TCP_TIMEOUT) {
                std::cerr << "[TCP] Input connection time out" << std::endl;
                ok_ = false;
                return 0;
            }

            if (Sock_read(static_cast<Sock*> (transport_), &tcp_size, sizeof(uint32_t),
                          NULL,MSG_DONTWAIT | MSG_PEEK) < 0) {
                if (errno== EAGAIN)
                    return 0;
                else {
                    ok_ = false;
                    return 0;
                }
            }

            size = Sock_read(static_cast<Sock*> (transport_), &tcp_size, sizeof(uint32_t),
                             NULL,MSG_WAITALL);
            if (static_cast<unsigned int> (size) < sizeof(uint32_t)) {
                ok_ = false;
                return 0;
            }

            tcp_size = ntohl(tcp_size);
            if (tcp_size > rtp::MAX_PAYLOAD_SIZE) {
                std::cerr << "[TCP] Received a too big packet, what's wrong? " << tcp_size
                    << std::endl;
                ok_ = false;
                return 0;
            }

            size = Sock_read(static_cast<Sock*> (transport_), received_packet_.getBuffer(),
                             tcp_size, NULL,MSG_WAITALL);
            if (size < tcp_size) {
                std::cerr << "[TCP] Warning Received only " << size << "/" << tcp_size << " bytes"
                    << std::endl;
                ok_ = false;
                return 0;
            }

            last_pkt_time_ = time(0);
#if ENABLE_DEBUG
            std::cerr << "[TCP] Received " << size << " bytes" << std::endl;
#endif
            received_packet_.parseBuffer(size);
            return size;
        }

        rtp::Packet &TCPInput::getPacket()
        {
            return received_packet_;
        }

        bool TCPInput::Check(std::string const &uri)
        {
            return (uri.find("tcp://") == 0);
        }

        Input *TCPInput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("tcp://") + 6);
            return new TCPInput(new_uri.c_str());
        }

        void TCPOutput::parse_uri(const std::string &uri, std::string &host, std::string &port)
        {
            std::string::size_type pos = uri.find_last_of(':');
            host = uri.substr(0, pos);

            if (pos == std::string::npos)
                throw std::runtime_error(
                                         "You must specify an host and destination port for TCP Output");

            port = uri.substr(++pos);
        }

        TCPOutput::TCPOutput(char const *uri) :
            Output(uri), transport_(0), ok_(false)
        {
            recover();
        }

        TCPOutput::~TCPOutput()
        {
            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));
        }

        bool TCPOutput::isAvailable()
        {
            return ok_;
        }

        bool TCPOutput::recover()
        {
            ok_ = false;

            std::string host, port;
            parse_uri(uri_, host, port);

            if (!is_a_port(port))
                throw std::runtime_error(
                                         "Please specify a valid even destination port for TCP Output");

            if (transport_) {
                Sock_close(static_cast<Sock*> (transport_));
                transport_ = 0;
            }

            while (!(transport_ = Sock_connect(host.c_str(), port.c_str(), 0, TCP, 0)))
                sleep(1);
            set_socket_timeout(static_cast<Sock*> (transport_), SO_SNDTIMEO);
            std::cerr << "Up on: " << host << ":" << port << std::endl;

            ok_ = true;
            return ok_;
        }

        unsigned long TCPOutput::write(unsigned char *buffer, unsigned long buffer_size)
        {
            if (!buffer_size)
                return 0;

            uint32_t sent = htonl(buffer_size);
            char *tcp_buffer = (char*) malloc(buffer_size + sizeof(uint32_t));

            memcpy(tcp_buffer, &sent, sizeof(uint32_t));
            memcpy(tcp_buffer + sizeof(uint32_t), buffer, buffer_size);

            sent = Sock_write(static_cast<Sock*> (transport_), tcp_buffer, buffer_size
                + sizeof(uint32_t), NULL, 0);
            if (sent != buffer_size + sizeof(uint32_t)) {
                std::cerr << "Warning " << sent << "/" << buffer_size << " bytes sent" << std::endl;
                ok_ = false;
            }

            free(tcp_buffer);
            return sent;
        }

        bool TCPOutput::Check(std::string const &uri)
        {
            return (uri.find("tcp://") == 0);
        }

        Output *TCPOutput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("tcp://") + 6);
            return new TCPOutput(new_uri.c_str());
        }
    }
}

#ifdef TEST
#include <iostream>
#include <unistd.h>

int main()
{
    flux::rtp::Packet p;
    p.setVersion(2);
    p.setSSRC(12345);
    p.regenBuffer();

    if (fork() == 0) {
        flux::io::TCPInput in("9000");

        try {
            flux::io::TCPOutput test1("localhost");
            std::cerr << "TCPOutput allocated without port" << std::endl;
            return 1;
        }
        catch(std::runtime_error &e) {}

        try {
            flux::io::TCPOutput test1("localhost:");
            std::cerr << "TCPOutput allocated with empty port" << std::endl;
            return 2;
        }
        catch(std::runtime_error &e) {}

        try {
            flux::io::TCPOutput test1("localhost:abc");
            std::cerr << "TCPOutput allocated with invalid port" << std::endl;
            return 3;
        }
        catch(std::runtime_error &e) {}

        try {
            flux::io::TCPOutput test1("localhost:9001");
            std::cerr << "TCPOutput allocated with odd port" << std::endl;
            return 4;
        }
        catch(std::runtime_error &e) {}

        in.read();
        flux::rtp::Packet &np = in.getPacket();
        //std::cout << "Packet Version " << np.getVersion() << " == " << p.getVersion() << std::endl;
        //std::cout << "Packet SSRC " << np.getSSRC() << " == " << p.getSSRC() << std::endl;

        if (np.getVersion() != p.getVersion()) {
            std::cerr << "Received packet version differs from original packet version" << std::endl;
            return 5;
        }

        if (np.getSSRC() != p.getSSRC()) {
            std::cerr << "Received packet SSRC differs from original packet SSRC" << std::endl;
            return 6;
        }

        std::cout << "+ TCP I/O TEST SUITE PASSED" << std::endl;
        return 0;
    }
    else {
        sleep(1);
        flux::io::TCPOutput out("localhost:9000");
        out.write(p.getBuffer(), p.getBufferSize());
        return 0;
    }
}
#endif

