/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include <cstring>

#include <stdexcept>
#if ENABLE_DEBUG
# include <iostream>
#endif

extern "C"
{
#include <netembryo/wsocket.h>
}

#include "io_udp.hpp"

namespace flux
{
    namespace io
    {
        UDPInput::UDPInput(char const *uri) :
            Input(uri), transport_(0), ok_(false)
        {
            recover();
        }

        UDPInput::~UDPInput()
        {
            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));
        }

        bool UDPInput::isAvailable()
        {
            return ok_;
        }

        bool UDPInput::recover()
        {
            ok_ = false;
            fd_ = -1;

            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));

            if (!is_a_port(uri_))
                throw std::invalid_argument("UDPInput requires a valid even port number as parameter");

            transport_ = Sock_bind(NULL, uri_.c_str(), NULL, UDP, 0);
            if (!transport_)
                throw std::runtime_error("UDPInput could not create server");

            fd_ = Sock_fd(static_cast<Sock*>(transport_));
            ok_ = true;
            return ok_;
        }

        void UDPInput::feed()
        {
            received_packet_.parseBuffer(0);
        }

        unsigned long UDPInput::read()
        {
            received_packet_.parseBuffer(0);

            if (!ok_)
                return 0;

            long size = Sock_read(static_cast<Sock*> (transport_), received_packet_.getBuffer(),
                                  rtp::MAX_PAYLOAD_SIZE, NULL,MSG_DONTWAIT);
            if (size < 0) {
                if (errno== EAGAIN)
                    return 0;
                else {
                    ok_ = false;
                    return 0;
                }
            }

#if ENABLE_DEBUG
            std::cerr << "[UDP] Received " << size << " bytes" << std::endl;
#endif
            received_packet_.parseBuffer(size);
            return size;
        }

        rtp::Packet &UDPInput::getPacket()
        {
            return received_packet_;
        }

        bool UDPInput::Check(std::string const &uri)
        {
            return (uri.find("udp://") == 0);
        }

        Input *UDPInput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("udp://") + 6);
            return new UDPInput(new_uri.c_str());
        }

        void UDPOutput::parse_uri(const std::string &uri, std::string &host, std::string &port)
        {
            std::string::size_type pos = uri.find_last_of(':');
            host = uri.substr(0, pos);

            if (pos == std::string::npos)
                throw std::runtime_error("You must specify an host and destination port for UDP Output");
            port = uri.substr(++pos);
        }

        UDPOutput::UDPOutput(char const *uri) :
            Output(uri), transport_(0), ok_(false)
        {
            recover();
        }

        UDPOutput::~UDPOutput()
        {
            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));
        }

        bool UDPOutput::isAvailable()
        {
            return ok_;
        }

        bool UDPOutput::recover()
        {
            ok_ = false;

            std::string host, port;
            parse_uri(uri_, host, port);

            if (!is_a_port(port))
                throw std::runtime_error("Please specify a valid even destination port for UDP Output");

            if (transport_)
                Sock_close(static_cast<Sock*> (transport_));

            //printf("UDP Output to %s:%s\n", host.c_str(), port.c_str());
            transport_ = Sock_connect(host.c_str(), port.c_str(), 0, UDP, 0);
            if (!transport_)
                throw std::runtime_error("UDP Output destination unreachable");

            ok_ = true;
            return ok_;
        }

        unsigned long UDPOutput::write(unsigned char *buffer, unsigned long buffer_size)
        {
            if (!buffer_size)
                return 0;

            unsigned long sent = Sock_write(static_cast<Sock*> (transport_), buffer, buffer_size,
                                            0, 0);
            if (sent < buffer_size)
                fprintf(stderr, "(WARN) UDPOutput sent %lu/%lu bytes\n", sent, buffer_size);
            return sent;
        }

        bool UDPOutput::Check(std::string const &uri)
        {
            return (uri.find("udp://") == 0);
        }

        Output *UDPOutput::Alloc(std::string const &uri)
        {
            std::string new_uri = uri.substr(uri.find("udp://") + 6);
            return new UDPOutput(new_uri.c_str());
        }
    }
}

#ifdef TEST
#include <iostream>

int main()
{
    flux::io::UDPInput in("9000");
    flux::io::UDPOutput out("localhost:9000");

    try {
        flux::io::UDPOutput test1("localhost");
        std::cerr << "UDPOutput allocated without port" << std::endl;
        return 1;
    }
    catch(std::runtime_error &e) {}

    try {
        flux::io::UDPOutput test1("localhost:");
        std::cerr << "UDPOutput allocated with empty port" << std::endl;
        return 2;
    }
    catch(std::runtime_error &e) {}

    try {
        flux::io::UDPOutput test1("localhost:abc");
        std::cerr << "UDPOutput allocated with invalid port" << std::endl;
        return 3;
    }
    catch(std::runtime_error &e) {}

    try {
        flux::io::UDPOutput test1("localhost:9001");
        std::cerr << "UDPOutput allocated with odd port" << std::endl;
        return 4;
    }
    catch(std::runtime_error &e) {}

    //Test packet transmission
    flux::rtp::Packet p;
    p.setVersion(2);
    p.setSSRC(12345);
    p.regenBuffer();

    out.write(p.getBuffer(), p.getBufferSize());
    in.read();
    flux::rtp::Packet &np = in.getPacket();
    //std::cout << "Packet Version " << np.getVersion() << " == " << p.getVersion() << std::endl;
    //std::cout << "Packet SSRC " << np.getSSRC() << " == " << p.getSSRC() << std::endl;

    if (np.getVersion() != p.getVersion()) {
        std::cerr << "Received packet version differs from original packet version" << std::endl;
        return 5;
    }

    if (np.getSSRC() != p.getSSRC()) {
        std::cerr << "Received packet SSRC differs from original packet SSRC" << std::endl;
        return 6;
    }

    std::cout << "+ UDP I/O TEST SUITE PASSED" << std::endl;
    return 0;
}
#endif

