/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include "live.hpp"

#include <cstdlib>
#include <iostream>
#include <stdexcept>
#include <algorithm>

#include <time.h>

#include "../utils/mq_metadata.hpp"

#ifdef TRANSCODE
#   include "../utils/ffmpeg.hpp"
#   include "../rtp/pack/packetizer.hpp"
#   include "../rtp/pack/factory.hpp"
#endif

#define SEQ_TRESHOLD 5

namespace flux
{
    namespace parse
    {
        LiveParser::LiveParser() :
            Parser(), ssrc_(rtp::Packet::GenerateSsrc()),
            last_timestamp_(rtp::Packet::CommonTimestamp()), prev_ts_(0), last_seq_(-1)
        {

        }

        LiveParser::~LiveParser()
        {
#if ENABLE_DEBUG
            std::cerr << "DELETING PARSER " << this << std::endl;
#endif

        }

        Parser *LiveParser::Alloc(std::string const &opts)
        {
            return new LiveParser();
        }

        bool LiveParser::parse(rtp::Packet const &p)
        {
            p_.parseBuffer(0);

            if (!p.getBufferSize())
                return false;

            if ((last_seq_ >= 0) && p.getSeq() == static_cast<unsigned short>(last_seq_)) {
                std::cerr << "[Live:" << p.getPayloadType() << "] Duplicated packet" << std::endl;
                return false;
            }

            last_seq_ = static_cast<int>(p.getSeq());
            last_timestamp_ += (p.getTimestamp() - prev_ts_);
            prev_ts_ = p.getTimestamp();

            p_ = p;
            p_.setSSRC(ssrc_);
            p_.setTimestamp(last_timestamp_);
            p_.regenBuffer();

            return true;
        }

        unsigned char *LiveParser::getResult()
        {
            return p_.getBuffer();
        }

        unsigned long LiveParser::getResultSize()
        {
            return p_.getBufferSize();
        }

        bool LiveParser::hasMore()
        {
            return false;
        }

#ifdef TRANSCODE
        EmptyLiveParser::EmptyLiveParser() :
            LiveParser(), last_valid_input_time_(0), fake_seq_(0), frame_num_(0)
        {
            FFmpeg ffmpeg = FFmpeg();
            AVFrame *black_frame = FFmpeg::MakeBlackFrame(640, 480);
            unsigned char buffer[32768];
            unsigned long encoded_buffer_size;
            std::list<rtp::Packet*> result_buffers;

            if (!ffmpeg.initEncodeContext(FFmpeg::CODEC_ID_MPEG2VIDEO, 640, 480)) {
                ffmpeg.FreeFrame(black_frame);
                throw std::runtime_error("Failed to initialize encoder for black frames");
            }

            encoded_buffer_size = ffmpeg.encode(black_frame, buffer, 32768);

            rtp::Packetizer * packer = rtp::PacketizerFactory::Create(ssrc_, 0, "MPV");
            if (packer->parse(result_buffers, buffer, encoded_buffer_size, NULL, 0) == rtp::Packetizer::Ok)
                black_pkts_ = result_buffers;

            ffmpeg.FreeFrame(black_frame);
        }

        EmptyLiveParser::~EmptyLiveParser()
        {

        }

        bool EmptyLiveParser::parse(rtp::Packet const &p)
        {
            if (!p.getBufferSize()) {
                pkts_to_send_.clear();

                if (black_pkts_.empty())
                    return false;

                if (time(NULL) - last_valid_input_time_ >= 3) {
                    pkts_to_send_ = black_pkts_;
#if ENABLE_DEBUG
                    std::cerr << "New black frame" << std::endl;
#endif
                    for(std::list<rtp::Packet*>::iterator p = pkts_to_send_.begin(); p != pkts_to_send_.end(); ++p) {
                        (*p)->setSSRC(ssrc_);
                        (*p)->setTimestamp(rtp::Packet::TimestampNow());
                        (*p)->setSeq(fake_seq_++);
                        (*p)->regenBuffer();
                    }
                    return true;
                }
                else
                    return false;
            }
            else
                last_valid_input_time_ = time(NULL);

            return LiveParser::parse(p);
        }

        unsigned char *EmptyLiveParser::getResult()
        {
            return pkts_to_send_.empty() ? 0 : pkts_to_send_.front()->getBuffer();
        }

        unsigned long EmptyLiveParser::getResultSize()
        {
            return pkts_to_send_.empty() ? 0 : pkts_to_send_.front()->getBufferSize();
        }

        bool EmptyLiveParser::hasMore() {
#if ENABLE_DEBUG
            std::cerr << "Seeding..." << std::endl;
#endif
            if (!pkts_to_send_.empty())
                pkts_to_send_.pop_front();
            return (!pkts_to_send_.empty());
        }

        Parser *EmptyLiveParser::Alloc(std::string const &opts)
        {
            return new EmptyLiveParser();
        }
#endif

    }
}

#ifdef TEST
#include <iostream>
#include <cstring>
#include <sys/stat.h>

int main()
{
    flux::parse::CopyParser parser;
    flux::rtp::Packet p;
    struct stat file_info;

    stat("test.rtp", &file_info);
    FILE * f = fopen("test.rtp", "r");
    fread(p.getBuffer(), 1, file_info.st_size, f);
    fclose(f);

    p.parseBuffer(file_info.st_size);
    parser.parse(p);

    if (p.getBufferSize() != parser.getResultSize()) {
        std::cerr << "Copied packet size differs from original packet size " << p.getBufferSize() << " != " << parser.getResultSize() << std::endl;
        return 1;
    }

    if (memcmp(p.getBuffer(), parser.getResult(), parser.getResultSize()) != 0) {
        std::cerr << "Copied packet differs from original packet" << std::endl;
        return 2;
    }

    std::cout << "+ COPYPARSER TEST SUITE PASSED" << std::endl;
    return 0;
}

#endif
