/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include "../rtp/unpack/depacketizer.hpp"
#include "../rtp/unpack/factory.hpp"

#include "../rtp/pack/packetizer.hpp"
#include "../rtp/pack/factory.hpp"

#include "../rtp/packet.hpp"

#include "transcode.hpp"

#include <iostream>
#include <cstring>
#include <cstdlib>

namespace flux
{
  namespace parse
  {
    TranscodeParser::TranscodeParser(bool repack, unsigned int ow, unsigned int oh)
    : Parser(), depacker_(0), repack_(repack), repacker_(0),
      tmp_buffer_size_(0), output_width_(ow), output_height_(oh),
      ssrc_(rtp::Packet::GenerateSsrc())
    {
    }

    TranscodeParser::~TranscodeParser()
    {
      delete depacker_;
    }

    Parser *TranscodeParser::Alloc(std::string const &opts)
    {
      unsigned int width = 0;
      unsigned int height = 0;

      if (!opts.empty()) {
        std::string::size_type sep = opts.find("x");
        if (sep != std::string::npos) {
          width = atoi(opts.substr(0, sep).c_str());
          height = atoi(opts.substr(sep+1).c_str());
        }
      }

      return new TranscodeParser(false, width, height);
    }

    Parser *TranscodeParser::PackAlloc(std::string const &opts)
    {
        TranscodeParser *p = dynamic_cast<TranscodeParser*>(Alloc(opts));
        p->repack_ = true;
        return p;
    }

    bool TranscodeParser::transcode(void * fbp, void * cbp)
    {
        rtp::Depacketizer::FrameBuffer *fb = static_cast<rtp::Depacketizer::FrameBuffer *>(fbp);
        //rtp::Depacketizer::ConfigBuffer *cb = static_cast<rtp::Depacketizer::ConfigBuffer *>(cbp);

        AVFrame *iframe = ffmpeg_.decode(fb->data, fb->len);
        if (!iframe)
            return false;

        if (!ffmpeg_.initScaleContext(output_width_, output_height_)) {
            std::cerr << "Failed to initialize scaling support" << std::endl;
            return false;
        }

        iframe = ffmpeg_.scaleFrame(iframe);
        if (!iframe)
            return false;

        if (!ffmpeg_.initEncodeContext(FFmpeg::CODEC_ID_MPEG2VIDEO, output_width_, output_height_)) {
            std::cerr << "Failed to initialize encoder" << std::endl;
            return false;
        }

        tmp_buffer_size_ = ffmpeg_.encode(iframe, tmp_buffer_, MAX_BUFFER_SIZE);
        if (!tmp_buffer_size_)
            return false;

        #if ENABLE_DEBUG
        std::cerr << "[Transcode] Frame of Size: " << output_width_ << "x" << output_height_ << " (" << tmp_buffer_size_ << ")" << std::endl;
        #endif

        return true;
    }

    bool TranscodeParser::parse(rtp::Packet const &p)
    {
        bool res = false;
        tmp_buffer_size_ = 0;

        if (!p.getBufferSize())
            return false;

        if (!depacker_) {
            depacker_ = rtp::DepacketizerFactory::Create(p.getPayloadType());
            if (!depacker_)
                return res;

            if (depacker_->getName() == "MPV") {
                if (!ffmpeg_.initDecodeContext(FFmpeg::CODEC_ID_MPEG2VIDEO)) {
                    std::cerr << "Unable to initialize decoder in transcoding" << std::endl;
                    delete depacker_;
                    depacker_ = 0;
                    return res;
                }
            }
            else {
                std::cerr << "Unsupported input codec in transcoding" << std::endl;
                delete depacker_;
                depacker_ = 0;
                return res;
            }
        }

        if (p.getPayloadType() != depacker_->getPayloadType()) {
            std::cerr << "Payload Type changed! " << depacker_->getPayloadType() << " expected, got " << p.getPayloadType() << std::endl;
            return res;
        }

        rtp::Depacketizer::FrameBuffer fb;
        rtp::Depacketizer::ConfigBuffer cb;
        rtp::Depacketizer::ParseResult result = depacker_->parse(p, &fb, &cb);

        if (result == rtp::Depacketizer::Ok)
            res = transcode(&fb, &cb);

        if (!res)
            return false;

        if (repack_) {
            res = false;
            if (!repacker_)
                repacker_ = rtp::PacketizerFactory::Create(ssrc_, p.getPayloadType());

            if (repacker_) {
                if (repacker_->parse(result_buffers_, tmp_buffer_, tmp_buffer_size_, NULL, 0) == rtp::Packetizer::Ok) {
                    #if ENABLE_DEBUG
                    std::cerr << "[Transcode] Repacked " << result_buffers_.size() << " packets" << std::endl;
                    #endif
                }
                else
                    std::cerr << "Failed to repack after transcoding" << std::endl;
            }
        }

        return res;
    }

    unsigned char *TranscodeParser::getResult()
    {
        return result_buffers_.empty() ? tmp_buffer_ : result_buffers_.front()->getBuffer();
    }

    unsigned long TranscodeParser::getResultSize()
    {
        return result_buffers_.empty() ? tmp_buffer_size_ : result_buffers_.front()->getBufferSize();
    }

    bool TranscodeParser::hasMore() {
        if (!result_buffers_.empty()) {
            rtp::Packet *p = result_buffers_.front();
            delete p;
            result_buffers_.pop_front();
        }
        return (!result_buffers_.empty());
    }
  }
}

