/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include <stdexcept>
#include <cstring>

#include "packet.hpp"

#include <sys/time.h>

namespace flux
{
  namespace rtp
  {
    unsigned long Packet::TimestampBase_ = 0;
    unsigned long Packet::SsrcLast_ = SsrcBase_ + 1;
    unsigned long Packet::GenerateUniqueSsrc()
    {
      return ++SsrcLast_;
    }

    unsigned int Packet::TimestampNow()
    {
      timeval tv;
      gettimeofday(&tv, NULL);
      unsigned int ts = (90000*tv.tv_sec);
      ts += (unsigned int)((2.0*90000*tv.tv_usec + 1000000.0)/2000000);
      return ts;
    }

    Packet::Packet()
    : version(2), padding(0), extension(0), marker(0), ssrc(0),
      csrc_count(0), payload_type(0), seq(0), time(0), data_size(12),
      payload_(data+data_size), payload_size_(0)
    {
    }

    Packet::~Packet()
    {}

    Packet &Packet::operator=(Packet const&other)
    {
      version = other.version;
      padding = other.padding;
      extension = other.extension;
      marker = other.marker;
      ssrc = other.ssrc;
      csrc_count = other.csrc_count;
      payload_type = other.payload_type;
      seq = other.seq;
      time = other.time;
      data_size = other.data_size;

      for(int i=0; i < csrc_count; ++i)
        csrc[i] = other.csrc[i];

      memcpy(data, other.data, data_size);

      payload_size_ = other.payload_size_;
      payload_ = data + getHeaderSize();

      return *this;
    }

    unsigned long Packet::getCSRC(unsigned short index) const
    {
      if (index >= getCSRCCount())
        throw std::out_of_range("Requested Contributing Source out of range");

      return csrc[index];
    }

    Packet &Packet::setCSRC(unsigned short index, unsigned long v)
    {
      if (index >= getCSRCCount())
        throw std::out_of_range("Trying to set a Contributing Source greater then the number of registered contribution sources");

      csrc[index] = v;
      return *this;
    }

    Packet &Packet::addCSRC(unsigned long v)
    {
      if (getCSRCCount() >= MAX_CSRC_COUNT)
        throw std::overflow_error("Too many Contributing Sources registered, maximum allowed number is 15");

      csrc[csrc_count++] = v;
      return *this;
    }

    Packet &Packet::parseBuffer(unsigned short size)
    {
      if (size > MAX_PAYLOAD_SIZE)
        throw std::overflow_error("Packet size greater than maximum buffer size");
      data_size = size;

      if (data_size < 12)
        return *this;

      version = (data[0] >> 6);
      padding = (data[0] >> 5) & 0x1;
      extension = (data[0] >> 4) & 0x1;
      csrc_count = data[0] & 0xF;

      marker = data[1] >> 7;
      payload_type = data[1] & 0x7F;

      seq =  ((unsigned)data[2] << 8) | ((unsigned)data[3]);
      time = ((unsigned)data[4] << 24) | ((unsigned)data[5] << 16) | ((unsigned)data[6] << 8) | ((unsigned)data[7]);
      ssrc = ((unsigned)data[8] << 24) | ((unsigned)data[9] << 16) | ((unsigned)data[10] << 8) | ((unsigned)data[11]);

      for(int i=0; i<csrc_count; ++i) {
        int csrc_begin = i*4;
        csrc[i] = ((unsigned)data[12 + csrc_begin] << 24) | ((unsigned)data[12 + csrc_begin + 1] << 16) |
                  ((unsigned)data[12 + csrc_begin + 2] << 8) | ((unsigned)data[12 + csrc_begin + 3]);
      }

      payload_ = data + 12 + (csrc_count*4);
      payload_size_ = getBufferSize() - getHeaderSize();

      return *this;
    }

    Packet &Packet::regenBuffer()
    {
      data[0] = version << 6 | padding << 5 | extension << 4 | csrc_count;
      data[1] = marker << 7 | payload_type;

      data[2] = (seq >> 8) & 0xFF;
      data[3] = seq & 0xFF;

      data[4] = (time >> 24) & 0xFF;
      data[5] = (time >> 16) & 0xFF;
      data[6] = (time >> 8) & 0xFF;
      data[7] = time & 0xFF;

      data[8] = (ssrc >> 24) & 0xFF;
      data[9] = (ssrc >> 16) & 0xFF;
      data[10] = (ssrc >> 8) & 0xFF;
      data[11] = ssrc & 0xFF;

      unsigned long new_data_size = getHeaderSize()+getPayloadSize();
      unsigned char *new_payload_end = data + new_data_size;

      unsigned char *cur_payload_begin = payload();
      unsigned char *cur_payload_end = cur_payload_begin + getPayloadSize();

      if (new_payload_end != cur_payload_end) {
      	while(cur_payload_end >= cur_payload_begin)
        	*(new_payload_end--) = *(cur_payload_end--);

      	payload_ = data + getHeaderSize();
      	payload_size_ = getBufferSize() - getHeaderSize();
      }

      data_size = new_data_size;

      for(int i=0; i<csrc_count; ++i) {
        int csrc_begin = i*4;

        data[12 + csrc_begin] = (csrc[i] >> 24) & 0xFF;
        data[12 + csrc_begin + 1] = (csrc[i] >> 16) & 0xFF;
        data[12 + csrc_begin + 2] = (csrc[i] >> 8) & 0xFF;
        data[12 + csrc_begin + 3] = (csrc[i]) & 0xFF;
      }

      return *this;
    }
  }
}





#if TEST
#include <iostream>
#include <cstdio>
#include <cstring>
#include <sys/stat.h>

int main(int argc, char *argv[])
{
  struct stat file_info;
  flux::rtp::Packet p;
  unsigned char clone_buffer[flux::rtp::MAX_PAYLOAD_SIZE];

  stat("test.rtp", &file_info);
  FILE * f = fopen("test.rtp", "r");
  fread(p.getBuffer(), 1, file_info.st_size, f);
  fclose(f);

  p.parseBuffer(file_info.st_size);
  memcpy(clone_buffer, p.getBuffer(), p.getBufferSize());

  if ( p.getPayloadSize() != (p.getBufferSize() - p.getHeaderSize()) ) {
    std::cerr << "Payload size differs from expected value" << std::endl;
    return 6;
  }

  p.regenBuffer();
  if ( static_cast<unsigned long long>(file_info.st_size) != p.getBufferSize() ) {
    std::cerr << "Original packet size differs from regenerated packet size" << std::endl;
    return 1;
  }

  if ( memcmp(clone_buffer, p.getBuffer(), p.getBufferSize()) != 0 ) {
    std::cerr << "Original packet differs from regenerated packet" << std::endl;

    std::cerr << "Version: " << p.getVersion() << std::endl;
    std::cerr << "Padding: " << p.hasPadding() << std::endl;
    std::cerr << "Extension: " << p.hasExtension() << std::endl;
    std::cerr << "Marker: " << p.hasMarker() << std::endl;
    std::cerr << "SSRC: " << p.getSSRC() << std::endl;
    std::cerr << "Contributing Sources: " << p.getCSRCCount() << std::endl;
    for (int i=0; i<p.getCSRCCount(); ++i)
      std::cerr << "CSRC " << i << " : " << p.getCSRC(i) << std::endl;
    std::cerr << "Payload Type: " << p.getPayloadType() << std::endl;
    std::cerr << "Sequence Number: " << p.getSeq() << std::endl;
    std::cerr << "Timestamp: " << p.getTimestamp() << std::endl;
    std::cerr << "Payload Size: " << p.getPayloadSize() << std::endl;

    return 2;
  }

  p.setVersion(3);
  p.regenBuffer();
  p.parseBuffer(p.getBufferSize());

  if ( p.getVersion() != 3 ) {
    std::cerr << "Version of modified packet is not the expected one" << std::endl;
    return 3;
  }

  p.addCSRC(65535);
  p.regenBuffer();
  p.parseBuffer(p.getBufferSize());

  if ( p.getCSRCCount() != 1 ) {
    std::cerr << "Count of CSRC different from expected one" << std::endl;
    return 4;
  }

  if ( p.getBufferSize() != static_cast<unsigned long long>(file_info.st_size + 4) ) {
    std::cerr << "Unexpected packet size after adding a new csrc" << std::endl;
    return 5;
  }

  std::cout << "+ PACKET TEST SUITE PASSED" << std::endl;
  return 0;
}
#endif
