/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#if ENABLE_DEBUG
#include <cstdio>
#endif

#include <cstdlib>
#include <cstring>
#include <stdint.h>

#include "../packet.hpp"
#include "mpegvideo.hpp"

typedef struct {
#ifdef WORDS_BIGENDIAN
    uint32_t mbz:5;
    uint32_t t:1;
    uint32_t tr:10;
    uint32_t an:1;
    uint32_t n:1;
    uint32_t s:1;
    uint32_t b:1;
    uint32_t e:1;
    uint32_t p:3;
    uint32_t fbv:1;
    uint32_t bfc:3;
    uint32_t ffv:1;
    uint32_t ffc:3;
#else
    uint32_t tr_h:2;
    uint32_t t:1;
    uint32_t mbz:5;
    uint32_t tr_l:8;
    uint32_t p:3;
    uint32_t e:1;
    uint32_t b:1;
    uint32_t s:1;
    uint32_t n:1;
    uint32_t an:1;
    uint32_t ffc:3;
    uint32_t ffv:1;
    uint32_t bfc:3;
    uint32_t fbv:1;
#endif
    union {
        uint8_t data[1];
        struct {
            uint32_t ext_hdr;
            uint8_t data[1];
        } mpeg2;
    } pt;
} rtp_mpv_pkt;

#define RTP_MPV_PKT(pkt)        ((rtp_mpv_pkt *)(pkt.payload()))
#define RTP_MPV_DATA(pkt)        (RTP_MPV_PKT(pkt)->t ? RTP_MPV_PKT(pkt)->pt.mpeg2.data : RTP_MPV_PKT(pkt)->pt.data)
#define RTP_MPV_DATA_LEN(pkt)    (RTP_MPV_PKT(pkt)->t ? (pkt.getPayloadSize()-8) : (pkt.getPayloadSize()-4))

#ifdef WORDS_BIGENDIAN
#define RTP_MPV_TR(pkt)            (RTP_MPV_PKT(pkt)->tr)
#else
#define RTP_MPV_TR(pkt)            (RTP_MPV_PKT(pkt)->tr_h << 8 | RTP_MPV_PKT(pkt)->tr_l)
#endif


namespace flux
{
  namespace rtp
  {
    MpegVideoDepacketizer::MpegVideoDepacketizer()
    : Depacketizer(32, "MPV"), data(0), len(0), data_size(0), timestamp(0)
    { }

    MpegVideoDepacketizer::~MpegVideoDepacketizer()
    {
        if(data)
            free(data);
    }

    Depacketizer::ParseResult MpegVideoDepacketizer::parse(Packet const &pkt, Depacketizer::FrameBuffer *fr,
                                                           Depacketizer::ConfigBuffer *config)
    {
        if (!pkt.getPayloadSize())
            return Depacketizer::PacketEmpty;

      #if ENABLE_DEBUG
        fprintf(stderr, "[MPV]: header: mbz:%u t:%u tr:%u an:%u n:%u s:%u b:%u e:%u p:%u fbv:%u bfc:%u ffv:%u ffc:%u\n",
            RTP_MPV_PKT(pkt)->mbz, RTP_MPV_PKT(pkt)->t,
            RTP_MPV_TR(pkt),
            RTP_MPV_PKT(pkt)->an, RTP_MPV_PKT(pkt)->n,
            RTP_MPV_PKT(pkt)->s, RTP_MPV_PKT(pkt)->b,
            RTP_MPV_PKT(pkt)->e, RTP_MPV_PKT(pkt)->p,
            RTP_MPV_PKT(pkt)->fbv, RTP_MPV_PKT(pkt)->bfc,
            RTP_MPV_PKT(pkt)->ffv, RTP_MPV_PKT(pkt)->ffc);
      #endif

        if (this->len && (pkt.getTimestamp() != this->timestamp)) {
            //incomplete packet without final fragment
            this->len = 0;
            return Depacketizer::MalformedPacket;
        }

        // discard pkt if it's fragmented and the first fragment was lost
        if (!this->len && !RTP_MPV_PKT(pkt)->b)
            return Depacketizer::LostFirstFragment;

        unsigned long long pkt_len = RTP_MPV_DATA_LEN(pkt);

        if (this->data_size < pkt_len + this->len) {
            if ( !(this->data = static_cast<unsigned char*>(realloc(this->data, pkt_len + this->len))) ) {
                return Depacketizer::OutOfMemory;
            }
            this->data_size = pkt_len + this->len;
        }

        memcpy(this->data + this->len, RTP_MPV_DATA(pkt), pkt_len);
        this->len += pkt_len;

        if (!pkt.hasMarker()) {
            this->timestamp = pkt.getTimestamp();
            return Depacketizer::NeedNextFragment;
        } else {
            fr->data = this->data;
            fr->len  = this->len;
            fr->timestamp = this->timestamp;
            this->len = 0;
        }

        return Depacketizer::Ok;
    }
  }
}

