/* *
 * This file is part of Flux
 *
 * Copyright (C) 2009 by LScube team <team@lscube.org>
 * See AUTHORS for more details
 *
 * flux is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * flux is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with flux; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 * */

#include "config.h"

#include <iostream>
#include <cstring>
#include "ffmpeg.hpp"

//Make FFmpeg happy without C99 stdint
#define INT64_C(a) ((int64_t)(a ## LL))

extern "C"
{
#include <libavcodec/avcodec.h>
#include <libswscale/swscale.h>
#include <libavcodec/opt.h>
}

namespace flux
{
    bool FFmpeg::init_done_ = false;
    int FFmpeg::Init()
    {
        if(!FFmpeg::init_done_) {
            avcodec_init();
            avcodec_register_all();
            FFmpeg::init_done_ = true;
            return 1;
        }

        return 0;
    }

    FFmpeg::FFmpeg()
    : root(FFmpeg::Init()),
      decode_context_(0), input_frame_(avcodec_alloc_frame()),
      encode_context_(0), output_frame_(avcodec_alloc_frame()),
      resize_context_(0)
    {
    }

    FFmpeg::~FFmpeg()
    {
        if (decode_context_) {
            avcodec_close(decode_context_);
            av_free(decode_context_);
        }

        if (encode_context_) {
            avcodec_close(encode_context_);
            av_free(encode_context_);
        }

        if (input_frame_)
            av_free(input_frame_);

        if (output_frame_)
            av_free(output_frame_);

        if (resize_context_)
            sws_freeContext(resize_context_);
    }

    int FFmpeg::initEncodeContext(FFmpeg::CodecId codec_id, unsigned int ow, unsigned int oh)
    {
        if(encode_context_)
            return 2;

        #if ENABLE_DEBUG
          std::cerr << "Initializing Encoder" << std::endl;
        #endif

        AVCodecContext *ocontext = avcodec_alloc_context();
        if (!ocontext)
            return 0;

        ocontext->codec_type = CODEC_TYPE_VIDEO;
        ocontext->width = ow;
        ocontext->height = oh;
        ocontext->gop_size = 12;
        ocontext->pix_fmt = PIX_FMT_YUV420P;
        ocontext->time_base.den = 25;
        ocontext->time_base.num = 1;
        ocontext->bit_rate = 200*1024;
        ocontext->has_b_frames = 0;
        ocontext->max_b_frames = 0;
        ocontext->me_method = 1;

        AVCodec *oCodec = avcodec_find_encoder(static_cast<CodecID>(codec_id));
        if ( avcodec_open(ocontext, oCodec) < 0 ) {
          std::cerr << "Unable to open output codec" << std::endl;
          return 0;
        }

        encode_context_ = ocontext;
        return 1;
    }

    unsigned long FFmpeg::encode(AVFrame *input, unsigned char *output, unsigned int output_size)
    {
        if (!input)
            return 0;

        input->pts =  AV_NOPTS_VALUE;
        return avcodec_encode_video(encode_context_, output, output_size, input);
    }

    int FFmpeg::initDecodeContext(FFmpeg::CodecId codec_id)
    {
        if(decode_context_)
            return 2;

        #if ENABLE_DEBUG
          std::cerr << "Initializing Decoder" << std::endl;
        #endif

        AVCodec *iCodec = avcodec_find_decoder(static_cast<CodecID>(codec_id));
        if (!iCodec) {
            std::cerr << "Unable to find input codec" << std::endl;
            return 0;
        }

        AVCodecContext *icontext = avcodec_alloc_context();
        if (!icontext)
            return 0;

        avcodec_get_context_defaults(icontext);
        if ( avcodec_open(icontext, iCodec) < 0 ) {
            std::cerr << "Unable to open input codec" << std::endl;
            return 0;
        }

        decode_context_ = icontext;
        return 1;
    }

    AVFrame *FFmpeg::decode(unsigned char *input, unsigned int input_size)
    {
        int got_pic;
        avcodec_decode_video(decode_context_, input_frame_, &got_pic, input, input_size);
        if (!got_pic)
            return 0;
        return input_frame_;
    }

    int FFmpeg::initScaleContext(unsigned int width, unsigned int height)
    {
        if (resize_context_)
            return 2;

        if ( (width == 0) && (height == 0) ) {
            if (width == 0)
              width = decode_context_->width;
            if (height == 0)
              height = decode_context_->height;
        }

        if (!resize_context_) {
            avpicture_alloc(reinterpret_cast<AVPicture*>(output_frame_), PIX_FMT_YUV420P, width, height);
            resize_context_ = sws_getContext(decode_context_->width, decode_context_->height,  decode_context_->pix_fmt,
                                             width, height, PIX_FMT_YUV420P, SWS_FAST_BILINEAR, NULL, NULL, NULL);
        }

        if (!resize_context_) {
            std::cerr << "Invalid Conversion Context" << std::endl;
            return 0;
        }

        return 1;
    }

    AVFrame *FFmpeg::scaleFrame(AVFrame *frame)
    {
        int r = sws_scale(resize_context_, frame->data, frame->linesize, 0, decode_context_->height,
                          output_frame_->data, output_frame_->linesize);
        if (r < 0)
            return 0;
        return output_frame_;
    }

    AVFrame *FFmpeg::MakeBlackFrame(unsigned int width, unsigned int height)
    {
        AVFrame *f = avcodec_alloc_frame();
        unsigned long size = avpicture_get_size(PIX_FMT_YUV420P, width, height);
        uint8_t *buffer = (uint8_t*)av_malloc(size*sizeof(uint8_t));

        memset(buffer, 0, size);
        avpicture_fill((AVPicture *)f, buffer, PIX_FMT_YUV420P, width, height);

        return f;
    }

    void FFmpeg::FreeFrame(AVFrame *f)
    {
        av_free(f->data[0]);
        av_free(f);
    }
}

